from __future__ import annotations
import enum
import re

from bigsmiles.config import Config


class AtomChirality(enum.Enum):
    none = ""
    R = "@"
    S = "@@"


def get_charge_number(symbol_text: str, charge_symbol: str) -> tuple[str, int]:
    count = symbol_text.count(charge_symbol)
    if count > 1:
        # format like Fe++
        return symbol_text[:-count], count

    index_positive_symbol = symbol_text.find(charge_symbol)
    if index_positive_symbol == len(symbol_text) - 1:
        # format like Fe+
        return symbol_text[:-1], count

    # format like Fe+2
    return symbol_text[:-2], int(symbol_text[-1])


def get_charge(symbol_text: str) -> tuple[str, int]:
    if "+" in symbol_text:
        return get_charge_number(symbol_text, "+")
    elif "-" in symbol_text:
        symbol_text, count = get_charge_number(symbol_text, "-")
        return symbol_text, -1 * count

    return symbol_text, 0


def get_chirality(symbol_text: str) -> tuple[str, AtomChirality]:
    result = re.search(r'@{1,2}', symbol_text)
    if result is None:
        return symbol_text, AtomChirality.none

    return symbol_text.replace("@", ""), AtomChirality(result.group())


def get_isotope(symbol_text: str) -> tuple[str, int | None]:
    if symbol_text[0].isdigit():
        isotope = re.search(r'\d+', symbol_text).group()
        return symbol_text.lstrip(isotope), int(isotope)

    return symbol_text, None


def get_hydrogens(symbol_text: str) -> tuple[str, int | None]:
    result = re.search(r'H[\d]?', symbol_text)
    if result is None:
        return symbol_text, None

    if result.group()[-1].isdigit():
        count = int(result.group()[-1])
    else:
        count = 1

    return symbol_text[:result.span()[0]] + symbol_text[result.span()[1]:], count


def process_atom_symbol(symbol_text: str) -> tuple[str, tuple[int] | None, AtomChirality, int, int | None, int | None]:
    """
    Process atom symbol into symbol, valence, chiral, charge, isotope, hydrogens

    Parameters
    ----------
    symbol_text: str
        atom symbol
        'C', '[235U]', '	[OH3+]'

    Returns
    -------
    symbol_text: str
        Atom symbol
    valance: tuple[int]
        Accepted valance electrons - From 'Config'
    chiral: AtomChirality
        chirality
    charge: int
        electron charge
    isotope: int
        isotope number
    hydrogens: int
        explict hydrogens

    """
    if "[" in symbol_text:
        symbol_text = symbol_text.replace('[', "").replace(']', '')
        symbol_text, isotope = get_isotope(symbol_text)
        symbol_text, chiral = get_chirality(symbol_text)
        symbol_text, hydrogens = get_hydrogens(symbol_text)
        symbol_text, charge = get_charge(symbol_text)
    else:
        charge = 0
        chiral = AtomChirality.none
        isotope = None
        hydrogens = None

    if symbol_text in Config.atoms_with_valence:
        valence = Config.atoms_with_valence[symbol_text]["valence"]
    else:
        valence = None

    return symbol_text, valence, chiral, charge, isotope, hydrogens


class Atom:
    __slots__ = ["id_", "valance_possible", "chiral", "charge", "isotope", "_hydrogens", "symbol", "valance",
                 "organic", "bonds"]
    _tree_print_label = True

    def __init__(self, symbol: str, id_: int = None):
        self.id_ = id_
        self.symbol, self.valance_possible, self.chiral, self.charge, self.isotope, self._hydrogens = \
            process_atom_symbol(symbol)
        self.valance = self.valance_possible[0] if self.valance_possible is not None else None
        self.organic = True if self.symbol in Config.organics else False
        self.bonds = []

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        return str(self) + "{" + str(self.id_) + "}"

    @property
    def bonds_available(self) -> int | None:
        if self.valance is None:
            return None
        return self.valance - self.number_of_bonds

    @property
    def number_of_bonds(self) -> int:
        hydrogen_count = 0
        if self._hydrogens is not None:
            hydrogen_count = self._hydrogens
        return sum((bond.type_.value for bond in self.bonds)) + hydrogen_count

    @property
    def ring_indexes(self) -> list[int]:
        ring_index = []
        for bond in self.bonds:
            if bond.ring is not None:
                ring_index.append(bond.ring)

        return ring_index

    def to_string(self, show_hydrogens: bool = False) -> str:
        """
        Construct Atom symbol

        Parameters
        ----------
        show_hydrogens: bool
            add implicit hydrogens to string

        Returns
        -------
        text: str
            atom symbol

        """
        text = self.symbol
        bracket_flag = False

        if self.symbol == "H":
            bracket_flag = True

        if self.chiral is not AtomChirality.none:
            text += self.chiral.value
            if self.bonds_available == 1:
                show_hydrogens = True
            bracket_flag = True

        if (show_hydrogens and self.bonds_available is not None and self.bonds_available > 0) or \
                self._hydrogens is not None:
            if self._hydrogens is not None:
                text += f"H{self._hydrogens if self._hydrogens > 1 else ''}"
            else:
                text += f"H{self.bonds_available if self.bonds_available > 1 else ''}"

            bracket_flag = True

        if self.charge > 0:
            text += f"+{self.charge if self.charge > 1 else ''}"
            bracket_flag = True
        elif self.charge < 0:
            text += f"{self.charge}"
            bracket_flag = True

        if self.isotope is not None:
            text = str(self.isotope) + text
            bracket_flag = True

        if bracket_flag:
            text = "[" + text + "]"

        return text + "".join((str(id_) for id_ in self.ring_indexes))


class BondType(enum.Enum):
    single = 1
    double = 2
    triple = 3


bond_mapping = {
    "": BondType.single,
    "=": BondType.double,
    "#": BondType.triple
}


class Bond:
    __slots__ = ["id_", "type_", "symbol", "atom1", "atom2", "ring"]
    _tree_print_label = True

    def __init__(self,
                 symbol: str,
                 atom1: Atom | BondDescriptor | StochasticObject,
                 atom2: Atom | BondDescriptor | StochasticObject | None = None,
                 id_: int = None,
                 ring: int = None
                 ):
        self.id_ = id_
        self.type_ = bond_mapping[symbol]
        self.symbol = symbol
        self.atom1 = atom1
        self.atom2 = atom2
        self.ring = ring

    def __str__(self):
        text = self.symbol
        if Config.color_output:
            return Config.colors(text, "Blue")

        return text

    def __repr__(self):
        bond_repr_symbols = {
            Atom: "A",
            BondDescriptor: "BD",
            StochasticObject: "SO",
        }
        text = self.symbol + "{" + str(self.id_) + f"|{bond_repr_symbols[type(self.atom1)]}{self.atom1.id_}-" \
                                                   f">{bond_repr_symbols[type(self.atom2)]}{self.atom2.id_}" + "}"
        if Config.color_output:
            return Config.colors(text, "Blue")

        return text

    def __iter__(self):
        return iter((self.atom1, self.atom2))


class BondDescriptorTypes(enum.Enum):
    Left = "<"
    Right = ">"
    Dollar = "$"
    Implicit = ""


def process_bonding_descriptor_symbol(symbol: str) -> tuple[str, int]:
    """
    Process bonding descriptor into symbol and index

    Parameters
    ----------
    symbol: str
        bonding descriptor symbol
        '[>1]'

    Returns
    -------
    symbol: str
        bonding descriptor symbol
        '<', '>', '$', ''
    index: int
        bonding descriptor index

    """
    symbol = symbol.replace("[", "").replace("]", "")
    if not symbol:
        return symbol, 0

    if symbol[-1].isdigit():
        return symbol[0], int(symbol[-1])

    return symbol, 0


class BondDescriptor:
    __slots__ = ["id_", "symbol", "type_", "index_", "bond"]
    _tree_print_label = True

    def __init__(self, symbol: str, id_: int = None):
        self.id_ = id_
        self.symbol, self.index_ = process_bonding_descriptor_symbol(symbol)
        self.type_ = BondDescriptorTypes(self.symbol)
        self.bond = None

    def __str__(self):
        if self.index_ == 0 and not Config.show_bond_descriptor_zero_index:
            index = ""
        else:
            index = self.index_
        text = "[" + self.symbol + str(index) + "]"
        if Config.color_output:
            return Config.colors(text, "Green")

        return text

    def __repr__(self):
        text = "[" + str(self.symbol) + "]" + "{" + str(self.id_) + "}"

        if Config.color_output:
            return Config.colors(text, "Green")

        return text


class Branch:
    _tree_print_label = False
    __slots__ = ["nodes", "id_", "parent"]

    def __init__(self, parent: BigSMILES | StochasticFragment | Branch, id_: int = None):
        self.nodes: list[Atom | Bond | Branch | StochasticObject | BondDescriptor] = []
        self.id_ = id_
        self.parent = parent

    def __str__(self):
        return "(" + "".join((str(node) for node in self.nodes)) + ")"

    def __repr__(self):
        return "({" + str(self.id_) + "} " + " ".join((repr(node) for node in self.nodes)) + " {" + str(self.id_) + \
               "})"

    @property
    def in_stochastic_object(self) -> bool:
        return self.parent.in_stochastic_object

    @property
    def root(self) -> BigSMILES:
        return self.parent.root


class StochasticFragment:
    _tree_print_label = False
    __slots__ = ["nodes", "id_", "parent"]

    def __init__(self, parent: StochasticObject, id_: int = None):
        self.nodes: list[Atom | Bond | BondDescriptor | Branch | StochasticObject] = []
        self.id_ = id_
        self.parent = parent

    def __str__(self):
        return "".join((str(node) for node in self.nodes))

    def __repr__(self):
        return " ".join((repr(node) for node in self.nodes))

    @property
    def in_stochastic_object(self) -> bool:
        return self.parent.in_stochastic_object

    @property
    def root(self) -> BigSMILES:
        return self.parent.root


class StochasticObject:
    _tree_print_label = False
    __slots__ = ["nodes", "id_", "parent", "end_group_left", "end_group_right", "bond_left", "bond_right"]

    def __init__(self, parent: BigSMILES | StochasticFragment | Branch, id_: int = None):
        self.nodes: list[StochasticFragment] = []
        self.end_group_left = None
        self.end_group_right = None
        self.id_ = id_
        self.bond_left = None
        self.bond_right = None
        self.parent = parent

    def __str__(self):
        if Config.color_output:
            fragments = Config.colors(",", "Red").join((str(node) for node in self.nodes))
            return Config.colors("{", "Red") + str(self.end_group_left) + fragments + str(self.end_group_right) + \
                   Config.colors("}", "Red")

        fragments = ",".join((str(node) for node in self.nodes))
        return "{" + str(self.end_group_left) + fragments + str(self.end_group_right) + "}"

    def __repr__(self):
        if Config.color_output:
            fragments = Config.colors(",", "Red").join((repr(node) for node in self.nodes))
            return Config.colors("{", "Red") + repr(self.end_group_left) + fragments + repr(self.end_group_right) + \
                   Config.colors("}", "Red")

        fragments = ",".join((repr(node) for node in self.nodes))
        return "{" + repr(self.end_group_left) + fragments + repr(self.end_group_right) + "}"

    @property
    def implicit_endgroups(self) -> bool:
        return False  # TODO:

    @property
    def in_stochastic_object(self) -> bool:
        return True

    @property
    def root(self) -> BigSMILES:
        return self.parent.root


class BigSMILES:
    _tree_print_label = False
    __slots__ = ["nodes", "atoms", "bonds", "rings", "input_text", "_tokens"]

    def __init__(self, input_text: str):
        self.nodes: list[Atom | Bond | StochasticObject | Branch] = []
        self.atoms: list[Atom] = []
        self.bonds: list[Bond] = []
        self.rings: list[Bond] = []

        # process input string
        self.input_text = input_text
        self._tokens = []

        from bigsmiles.create_parse_tree import create_parse_tree
        create_parse_tree(self)

    def __str__(self):
        return "".join((str(node) for node in self.nodes))

    def __repr__(self):
        return " ".join((repr(node) for node in self.nodes))

    def __getitem__(self, obj: int | slice) -> Atom | list[Atom]:
        return self.atoms[obj]

    def __iter__(self) -> list[Atom]:
        return self.atoms

    @property
    def in_stochastic_object(self) -> bool:
        return False

    @property
    def root(self) -> BigSMILES:
        return self

    def print_tree(self, show_object_label: bool = True, print_repr: bool = False):
        """
        prints a tree representation of the parsed bigsmiles

        Parameters
        ----------
        show_object_label: bool
            show object labels
        print_repr: bool
            use repr() instead of str()

        """
        from bigsmiles.tree_to_string import tree_to_string
        print(tree_to_string(self, show_object_label, print_repr))




from bigsmiles.tokenizer import Token, TokenKind, tokenize
from bigsmiles.bigsmiles_constructor import BigSMILESConstructor, States, BigSMILESError


def map_atom(constructor: BigSMILESConstructor, tokens: list[Token], token: Token):
    if constructor.state is States.start:
        constructor.add_atom(token.value)

    elif constructor.state in (States.atom, States.bond_descriptor, States.stochastic_object_end,
                               States.stochastic_fragment, States.branch_start):
        constructor.add_bond_atom_pair("", token.value)  # add single bond

    else:
        raise BigSMILESError("Something went wrong in 'map_atom'.")  # should not be hit


def map_bond(constructor: BigSMILESConstructor, tokens: list[Token], token: Token):
    try:
        next_token = tokens.pop(0)
    except IndexError:
        raise BigSMILESError("Bond can't be at the end of a BigSMILES string.")

    if next_token.kind in (TokenKind.Atom, TokenKind.AtomExtend):
        constructor.add_bond_atom_pair(token.value, next_token.value)
    elif next_token.kind is not TokenKind.BondDescriptor:
        constructor.add_bond_bonding_descriptor_pair(token.value, next_token.value)
    elif next_token.kind is not TokenKind.StochasticStart:
        map_stochastic_object_start(constructor, tokens, next_token, token)
    else:
        raise BigSMILESError(f"Bonds must be followed by an Atom, Bonding Descriptor, or Stocastic Object. "
                             f"Parse completed: {constructor.bigsmiles}, issue token: {token}")


def map_bond_descriptor(constructor: BigSMILESConstructor, tokens: list[Token], token: Token):
    if tokens[0].kind == TokenKind.StochasticEnd:
        constructor.close_stochastic_fragment()
        constructor.close_stochastic_object(token.value)
        tokens.pop(0)
        return

    if constructor.state is States.stochastic_fragment:
        # first StochasticFragment symbol
        constructor.add_bonding_descriptor(token.value)
        return

    if constructor.state is States.branch_start:
        # first Branch symbol
        # check to make sure the branch closes immediately
        if tokens[0].kind != TokenKind.BranchEnd:
            raise BigSMILESError("If Bonding Descriptors is the first Branch symbol, it can only be followed by "
                                 "Branch End.")
    constructor.add_bond_bonding_descriptor_pair("", token.value)


def map_branch_start(constructor: BigSMILESConstructor, tokens: list[Token], token: Token):
    constructor.open_branch()


def map_branch_end(constructor: BigSMILESConstructor, tokens: list[Token], token: Token):
    constructor.close_branch()


def map_ring(constructor: BigSMILESConstructor, tokens: list[Token], token: Token):
    if constructor.state is not States.atom:
        raise BigSMILESError(f"Ring number must follow atoms.")

    constructor.add_ring(int(token.value))


def map_stochastic_object_start(constructor: BigSMILESConstructor, tokens: list[Token], token: Token,
                                bond_token: Token = None):
    try:
        next_token = tokens.pop(0)
    except IndexError:
        raise BigSMILESError("Stochastic objects must begin with a bond descriptor (or implicit bonding descriptor).")

    if next_token.kind not in (TokenKind.ImplictEndGroup, TokenKind.BondDescriptor):
        raise BigSMILESError(f"Bonds must be followed by an Atom. Parse completed: "
                             f"{constructor.bigsmiles}, issue token: {token}")

    if constructor.state is States.start:
        constructor.open_stochastic_object(next_token.value)
    else:
        if bond_token is None:
            constructor.open_stochastic_object_with_bond("", next_token.value)
        else:
            constructor.open_stochastic_object_with_bond(bond_token.value, next_token.value)


def map_stochastic_object_end(constructor: BigSMILESConstructor, tokens: list[Token], token: Token) -> TokenKind:
    raise BigSMILESError("Stochastic objects should end with bonding descriptor (or implicit bonding description)")


def map_bond_seperator(constructor: BigSMILESConstructor, tokens: list[Token], token: Token):
    constructor.close_open_stochastic_fragment()


def NotImplementedFunc(*args, **kwargs):
    raise NotImplementedError


map_tokens = {
    TokenKind.Bond: map_bond,
    TokenKind.Atom: map_atom,
    TokenKind.Aromatic: map_atom,
    TokenKind.AtomExtend: map_atom,
    TokenKind.BranchStart: map_branch_start,
    TokenKind.BranchEnd: map_branch_end,
    TokenKind.Ring: map_ring,
    TokenKind.Ring2: map_ring,
    TokenKind.BondEZ: NotImplementedFunc,
    TokenKind.Mix: NotImplementedFunc,
    TokenKind.Rxn: NotImplementedFunc,
    TokenKind.BondDescriptor: map_bond_descriptor,
    TokenKind.StochasticSeperator: map_bond_seperator,
    TokenKind.StochasticStart: map_stochastic_object_start,
    TokenKind.StochasticEnd: map_stochastic_object_end,
    TokenKind.ImplictEndGroup: map_bond_descriptor,
    TokenKind.BondDescriptorLadder: NotImplementedFunc
}


def tokens_to_objects(constructor: BigSMILESConstructor, tokens: list[Token]) -> TokenKind | None:
    """

    Main loop for converting tokens into BigSMILES objects.

    Parameters
    ----------
    constructor: BigSMILESConstructor

    tokens:

    Returns
    -------

    """
    while tokens:
        token = tokens.pop(0)

        func = map_tokens[token.kind]
        result = func(constructor, tokens, token)
        if isinstance(result, TokenKind):
            return result

    return None


def create_parse_tree(bigsmiles):
    """

    Main function that turns BigSMILES string into a BigSMILES object.
    Constructs BigSMILES tree in the provided object.

    Parameters
    ----------
    bigsmiles: BigSMILES
        BigSMILES object with BigSMILES string added as 'input_text'

    """
    tokens = tokenize(bigsmiles.input_text)
    constructor = BigSMILESConstructor(bigsmiles)
    tokens_to_objects(constructor, tokens)



import enum
from functools import wraps

from bigsmiles.bigsmiles import Atom, Bond, BondDescriptor, Branch, StochasticFragment, StochasticObject, \
    BigSMILES


class BigSMILESError(Exception):
    def __init__(self, text: str, optional: str = None):
        if optional is not None:
            text += "(" + optional + ")"
        super().__init__(text)


class States(enum.Enum):
    start = enum.auto()
    atom = enum.auto()
    bond_descriptor = enum.auto()
    branch_start = enum.auto()
    branch = enum.auto()
    stochastic_fragment = enum.auto()
    stochastic_object_start = enum.auto()
    stochastic_object_end = enum.auto()
    stochastic_object = enum.auto()


def check_atom_for_making_bond(bond: Bond, atom: Atom):
    """
    Check to see if atom can accept bond.

    Parameters
    ----------
    bond: Bond
        Bond to add to Atom
    atom: Atom
        Atom to receive Bond

    """
    if atom.bonds_available is None:
        pass
    elif atom.bonds_available == 0:
        if atom.valance == atom.valance_possible[-1] or atom.valance_possible[-1] < atom.valance + bond.type_.value:
            raise BigSMILESError("Too many bonds trying to be made.", repr(atom))

        for val in bond.atom1.valance_possible:
            if val > atom.valance:
                atom.valance = val
                break

    atom.bonds.append(bond)


def add_bond_to_connected_objects(bond: Bond):
    """ Adds bonds to Atoms, and BondDescriptor. """
    for obj in bond:
        if isinstance(obj, Atom):
            check_atom_for_making_bond(bond, obj)
        elif isinstance(obj, BondDescriptor):
            obj.bond = bond


def in_stochastic_object(func):
    """ Decorator to ensure function call only occurs within stochastic object."""
    @wraps(func)
    def _in_stochastic_object(*args, **kwargs):
        if not args[0].stack[-1].in_stochastic_object:
            raise BigSMILESError("Must be in stochastic object.")
        return func(*args, **kwargs)

    return _in_stochastic_object


class BigSMILESConstructor:

    def __init__(self, obj: BigSMILES):
        self.bigsmiles = obj

        # setup id counters
        self._atom_counter = 0
        self._bond_counter = 0
        self._bond_descriptor = 0
        self._branch_counter = 0
        self._stochastic_fragment = 0
        self._stochastic_object = 0

        self.state = States.start
        self.stack: list[BigSMILES | StochasticObject | StochasticFragment | Branch] = [self.bigsmiles]

    def _get_atom_id(self) -> int:
        self._atom_counter += 1
        return self._atom_counter - 1

    def _get_bond_id(self) -> int:
        self._bond_counter += 1
        return self._bond_counter - 1

    def _get_bond_descriptor_id(self) -> int:
        self._bond_descriptor += 1
        return self._bond_descriptor - 1

    def _get_branch_id(self) -> int:
        self._branch_counter += 1
        return self._branch_counter - 1

    def _get_stochastic_fragment_id(self) -> int:
        self._stochastic_object += 1
        return self._stochastic_object - 1

    def _get_stochastic_object_id(self) -> int:
        self._stochastic_object += 1
        return self._stochastic_object - 1

    def add_ring(self, ring_id: int) -> Bond:
        # check if ring already exists
        for ring in self.bigsmiles.rings:
            if ring.ring == ring_id:
                if ring.atom2 is not None:
                    raise BigSMILESError("Ring already formed")
                ring.atom2 = self._get_prior(self.stack[-1], (Atom,))
                return ring

        # Make new ring
        bond = Bond("", self._get_prior(self.stack[-1], (Atom,)), None, self._get_bond_id(), ring_id)
        self.bigsmiles.nodes.append(bond)
        self.bigsmiles.bonds.append(bond)
        add_bond_to_connected_objects(bond)

        return bond

    def add_atom(self, symbol: str) -> Atom:
        atom = Atom(symbol, self._get_atom_id())
        self.stack[-1].nodes.append(atom)
        self.bigsmiles.atoms.append(atom)

        self.state = States.atom
        return atom

    def add_bond(self, bond_symbol: str, atom1: Atom | BondDescriptor, atom2: Atom | BondDescriptor | None) -> Bond:
        bond = Bond(bond_symbol, atom1, atom2, self._get_bond_id())
        self.stack[-1].nodes.append(bond)
        self.bigsmiles.bonds.append(bond)
        add_bond_to_connected_objects(bond)
        return bond

    @in_stochastic_object
    def add_bonding_descriptor(self, bd_symbol: str) -> BondDescriptor:
        """ [<], [>], [$], [$1], [>2], ... """
        bd = BondDescriptor(bd_symbol, self._get_bond_descriptor_id())
        self.stack[-1].nodes.append(bd)

        self.state = States.bond_descriptor
        return bd

    def add_bond_atom_pair(self, bond_symbol: str, atom_symbol: str) -> tuple[Bond, Atom]:
        atom = Atom(atom_symbol, self._get_atom_id())
        prior_atom = self._get_prior(self.stack[-1], (Atom, BondDescriptor, StochasticObject))
        bond = self.add_bond(bond_symbol, prior_atom, atom)
        self.stack[-1].nodes.append(atom)
        self.bigsmiles.atoms.append(atom)

        self.state = States.atom
        return bond, atom

    @in_stochastic_object
    def add_bond_bonding_descriptor_pair(self, bond_symbol: str, bd_symbol: str) -> tuple[Bond, BondDescriptor]:
        bd = BondDescriptor(bd_symbol, self._get_bond_descriptor_id())
        prior_atom = self._get_prior(self.stack[-1], (Atom, BondDescriptor, StochasticObject))
        bond = self.add_bond(bond_symbol, prior_atom, bd)
        self.stack[-1].nodes.append(bd)

        self.state = States.bond_descriptor
        return bond, bd

    def open_branch(self):
        if isinstance(self.stack[-1], Branch) and len(self.stack[-1].nodes) == 0:
            raise BigSMILESError("BigSMILES string or branch can't start with branch")

        branch = Branch(self.stack[-1], self._get_branch_id())
        self.stack[-1].nodes.append(branch)
        self.stack.append(branch)
        self.state = States.branch_start

    def close_branch(self):
        if not isinstance(self.stack[-1], Branch):
            raise BigSMILESError("Error closing branch. Possible issues: "
                                 "\n\tClosing a branch with another intermediate node started."
                                 "\n\tNo starting branch symbol.")
        self.stack.pop()

    def open_stochastic_object(self, bd_symbol: str) -> StochasticFragment:
        stoch_obj = StochasticObject(self.stack[-1], self._get_stochastic_object_id())
        bd = BondDescriptor(bd_symbol, self._get_bond_descriptor_id())
        stoch_obj.end_group_left = bd
        self.stack[-1].nodes.append(stoch_obj)
        self.stack.append(stoch_obj)

        # open stochastic_fragment
        stoch_frag = StochasticFragment(self.stack[-1], self._get_stochastic_fragment_id())
        self.stack[-1].nodes.append(stoch_frag)
        self.stack.append(stoch_frag)
        self.state = States.stochastic_fragment
        return stoch_frag

    def open_stochastic_object_with_bond(self, bond_symbol: str, bd_symbol: str) -> StochasticFragment:
        stoch_obj = StochasticObject(self.stack[-1], self._get_stochastic_object_id())
        bd = BondDescriptor(bd_symbol, self._get_bond_descriptor_id())
        stoch_obj.end_group_left = bd

        prior_atom = self._get_prior(self.stack[-1], (Atom, BondDescriptor, StochasticObject))
        bond = Bond(bond_symbol, prior_atom, stoch_obj, self._get_bond_id())
        stoch_obj.bond_left = bond
        self.stack[-1].nodes.append(bond)

        self.stack[-1].nodes.append(stoch_obj)
        self.stack.append(stoch_obj)

        return self.open_stochastic_fragment()

    def close_stochastic_object(self, bd_symbol: str):
        if not isinstance(self.stack[-1], StochasticObject):
            raise BigSMILESError("Error closing StochasticObject. Possible issues: "
                                 "\n\tClosing a StochasticObject with another intermediate node started."
                                 "\n\tNo starting StochasticObject symbol.")

        self.stack[-1].end_group_right = BondDescriptor(bd_symbol, self._get_bond_descriptor_id())
        self.stack.pop()

    @in_stochastic_object
    def open_stochastic_fragment(self) -> StochasticFragment:
        stoch_frag = StochasticFragment(self.stack[-1], self._get_stochastic_fragment_id())
        self.stack[-1].nodes.append(stoch_frag)
        self.stack.append(stoch_frag)
        self.state = States.stochastic_fragment
        return stoch_frag

    @in_stochastic_object
    def close_open_stochastic_fragment(self) -> StochasticFragment:
        if not isinstance(self.stack[-1], StochasticFragment):
            raise BigSMILESError("Stochastic seperator can only follow a stochastic fragments.")

        self.close_stochastic_fragment()
        return self.open_stochastic_fragment()

    @in_stochastic_object
    def close_stochastic_fragment(self):
        if not isinstance(self.stack[-1], StochasticFragment):
            raise BigSMILESError("Error StochasticFragment branch. Possible issues: "
                                 "\n\tClosing a branch with another intermediate node started.")

        self.stack.pop()

    def _get_prior(self, obj: BigSMILES | StochasticObject | StochasticFragment | Branch, types_: tuple,
                   flag: bool = False) -> Atom | None:
        if obj.nodes:
            if isinstance(obj.nodes[-1], types_):
                return obj.nodes[-1]
            if isinstance(obj.nodes[-1], Branch):
                if isinstance(obj.nodes[-2], types_):
                    return obj.nodes[-2]
                if isinstance(obj.nodes[-2], Branch):
                    if isinstance(obj.nodes[-3], types_):
                        return obj.nodes[-3]
        elif hasattr(obj, "parent") and not flag:
            return self._get_prior(self.stack[-2], types_, flag=True)

        return None



"""

Code for turing a python class into a terminal printable tree.

"""

from typing import Protocol


class TreeConfig:
    tree_symbol_options = {
        'ascii': ('|', '|-- ', '+-- '),
        'ascii-ex': ('\u2502', '\u251c\u2500\u2500', '\u2514\u2500\u2500'),
        'ascii-exr': ('\u2502', '\u251c\u2500\u2500', '\u2570\u2500\u2500'),
        'ascii-em': ('\u2551', '\u2560\u2550\u2550', '\u255a\u2550\u2550'),
        'ascii-emv': ('\u2551', '\u255f\u2500\u2500', '\u2559\u2500\u2500'),
        'ascii-emh': ('\u2502', '\u255e\u2550\u2550', '\u2558\u2550\u2550'),
    }
    symbols = 'ascii-ex'  # some computers may need to show different symbols
    tab = "    "
    space = " "

    @classmethod
    def vertical_line(cls) -> str:
        return cls.tree_symbol_options[cls.symbols][0]

    @classmethod
    def tee(cls) -> str:
        return cls.tree_symbol_options[cls.symbols][1]

    @classmethod
    def corner(cls) -> str:
        return cls.tree_symbol_options[cls.symbols][2]


class TreeNode(Protocol):
    _tree_print_label: bool


class TreeIntermediateNode(TreeNode):
    nodes: list


def to_repr(obj) -> str:
    return repr(obj)


def to_str(obj) -> str:
    return str(obj)


def tree_to_string(root_node: TreeIntermediateNode, show_object_labels: bool = True, print_repr: bool = False) -> str:
    """
    Creates a string representation of a tree for printing.

    Parameters
    ----------
    root_node: TreeIntermediateNode
        Root node
    show_object_labels: bool
        show object labels
    print_repr: bool
        print str or print repr of objects
        True: repr() is called
        False: str() is called

    Returns
    -------
    text: str
        String with the structure of a tree.

    """
    text = ""

    # add root node
    if show_object_labels:
        text = type(root_node).__name__ + ": "

    text += str(root_node)

    # add main part of tree
    text += tree_to_string_loop(root_node.nodes, [], show_object_labels, to_repr if print_repr else to_str)

    return text


def tree_to_string_loop(
        nodes: list[TreeNode, TreeIntermediateNode],
        spacers: list[bool],
        show_object_label: bool,
        func: callable
) -> str:
    """
    Creates the main body of the tree

    Parameters
    ----------
    nodes: list[TreeNode | TreeIntermediateNode]
        nodes
    spacers: list[bool]
        list records weather a continuation line (TreeConfig.vertical_line) is needed in the tree
        True: No line
        False: Add line
    show_object_label: bool
        show object/class type.__name__
    func: callable
        calls str() or repr()

    Returns
    -------
    text: str

    """
    text = ""

    for node in nodes:
        text += create_row(node, spacers, show_object_label, func, node == nodes[-1])

        # if interior node, do recursion
        if hasattr(node, "nodes"):
            text += tree_to_string_loop(
                node.nodes,
                spacers + [node == nodes[-1]],
                show_object_label,
                func
            )

    return text


def create_row(
        node: TreeNode | TreeIntermediateNode,
        spacers: list[bool],
        show_object_label: bool,
        func: callable,
        last_node: bool
) -> str:
    """
    Creates a row in the tree string representation.

    Parameters
    ----------
    node: TreeNode | TreeIntermediateNode
        node
    spacers: list[bool]
        list records weather a continuation line (TreeConfig.vertical_line) is needed in the tree
        True: No line
        False: Add line
    show_object_label: bool
        show object/class type.__name__
    func: callable
        calls str() or repr()
    last_node: bool
        last node in branch --> switch tee symbol to corner

    Returns
    -------
    text: str

    """
    line = "\n"

    # add spacer
    for i, spacer in enumerate(spacers):
        if not spacer:
            line += TreeConfig.vertical_line()
        line += TreeConfig.tab

    # add tree connection symbol
    if last_node:
        line += TreeConfig.corner() + TreeConfig.space
    else:
        line += TreeConfig.tee() + TreeConfig.space

    # add class label
    if show_object_label:
        line += type(node).__name__ + ": "

    # add node symbol
    if node._tree_print_label:
        line += str(node)
    else:
        line += func(node)

    return line



import pkg_resources  # part of setuptools
__version__ = pkg_resources.require("bigsmiles")[0].version

from bigsmiles.config import Config
from bigsmiles.tokenizer import BigSMILESTokenizeError
from bigsmiles.bigsmiles_constructor import BigSMILESError
from bigsmiles.bigsmiles import Atom, Bond, Branch, BondDescriptor, StochasticObject, StochasticFragment, BigSMILES, \
    AtomChirality, BondDescriptorTypes, BondType

errors = [
    BigSMILESError,
    BigSMILESTokenizeError
]

enums = [
    BondType,
    BondDescriptorTypes,
    AtomChirality
]
